/* Copyright (c) 2009, Delft University of Technology

    This file is part of the MEGS package.

    The MEGS pakcage is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    The MEGS pakcage is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the MEGS package.  If not, see
    <http://www.gnu.org/licenses/>.
*/
package engine;
import java.io.*; 
import java.util.StringTokenizer;

/**
 * @author menno
 * Reads the data from a single file, all events are assumed to be ordered
 * Takes usr3 output
 */
public class SingleUARTFileReader implements EventReader {
	
	
	private BufferedReader reader;
	
	
	/**
	 * The String representing the next event, this is needed to implement hasMoreEvents()
	 */
	private String nextString;

	/**
	 * Constructs a reader from file f
	 * @param f
	 */
	public SingleUARTFileReader(File f) throws FileNotFoundException
	{
		reader = new BufferedReader(new FileReader(f));
		loadEvent();
		
	}
	
	/* (non-Javadoc)
	 * @see EventReader#hasMoreEvents()
	 */
	public boolean hasMoreEvents() {
		return (nextString != null);
	}

	/* (non-Javadoc)
	 * @see EventReader#nextEvent()
	 */
	public Event nextEvent() {

		
		//keep trying until you find a real event 
		while(hasMoreEvents())
		{
			Event event = createEvent(nextString);
		
			loadEvent();
			
			if(event != null)
			{
				return event;
			}
		}
		
		//no events left
		return null;
	}
	
	/**
	 * Puts the next event into nextString
	 */
	private void loadEvent()
	{
		try
		{
			nextString = reader.readLine();
		}catch(IOException e)
		{
			System.out.println("Error reading next event: " + e);
		}
	}
	
	/**
	 * Takes a string and creates an event. String format is:
	 * String: Nodenumber - hours:minutes:seconds: Message
	 * <nodenum> <time (s)> (<time>) <file>  <type> <num> (<num>)
	 * Seconds is a double
	 * @param s
	 * @return
	 */
	public static Event createEvent(String s)
	{
		//System.out.println("S:" + s);
		//Ignore anytihg that doesn't start with a number
		if(!s.matches("^[0-9]* .*$"))
		{
			return null;
		}
		
		StringTokenizer tok = new StringTokenizer(s, " ");
		
		try
		{
			//parse the intro if not ignore the event		

			int node = Integer.parseInt(tok.nextToken());
			double time = 0;
			try
			{
				time = Double.parseDouble(tok.nextToken());
			}
			catch(NumberFormatException nfe)
			{
				//Sometimes there are rackRunner events in the parsed output
				//These don't have a timestamp
				return null;
			}
			
			//skip the second notion of time
			tok.nextToken();
			//skip the filename
			tok.nextToken();
			
			String eventType = tok.nextToken();
			if(eventType.startsWith("_ASSERT"))
			{
				AssertEvent e = new AssertEvent();
				//make an assert
				e.node = node;
				e.time = time;
				
				e.assertNum = Integer.parseInt(tok.nextToken());
								
				return e;
			}
			else if(eventType.startsWith("__STATUS") || eventType.startsWith("_STATUS"))
			{
				//make a status update
				StatusEvent e = new StatusEvent();
				
				e.node = node;
				e.time = time;
				if(eventType.startsWith("__STATUS"))
				{
					e.name = eventType.substring(9); //skip __STATUS_
				}
				else
				{
					e.name = eventType.substring(8); //skip _STATUS_
				}
				e.value = Integer.parseInt(tok.nextToken());
				return e;
			}
			else if(eventType.startsWith("BROADCAST_ADDR"))
			{
				//make a broadcast event
				BroadcastEvent e = new BroadcastEvent();
				
				e.node = node;
				e.time = time;
				return e;
			}
			else if(eventType.startsWith("_RECVSEND_ADDR"))
			{
				//make a recieve event
				ReceiveSendMsgEvent e = new ReceiveSendMsgEvent();
				
				e.node = node;
				e.time = time;
				e.fromNode = Integer.parseInt(tok.nextToken());
				return e;
			}
			else if(eventType.startsWith("_RECVBCST_ADDR"))
			{
				//make a recieve event
				ReceiveBroadcastMsgEvent e = new ReceiveBroadcastMsgEvent();
				
				e.node = node;
				e.time = time;
				e.fromNode = Integer.parseInt(tok.nextToken());
				return e;
			}
			else if(eventType.startsWith("_INTERCEPT_ADDR"))
			{
				//make a recieve event
				InterceptMsgEvent e = new InterceptMsgEvent();
				
				e.node = node;
				e.time = time;
				e.fromNode = Integer.parseInt(tok.nextToken());
				return e;
			}
			else if(eventType.startsWith("__SEND_ADDR"))
			{
				//make a send event
				SendSingleMsgEvent e = new SendSingleMsgEvent();
				
				e.node = node;
				e.time = time;
				e.toNode = Integer.parseInt(tok.nextToken());
			
				return e;
			}
			else
			{
				if(!Settings.IGNORE_UNKNOWN_EVENTS)
				{
					//System.out.println("Unrecognized event("+assertOrStatus+"): " + s);
					UnknownEvent e = new UnknownEvent();
					e.node = node;
					e.time = time;
					return e;
				}
			}

		}catch(Exception e)
		{
			System.out.println("Error creating event " + e + " on String: " + s);
		}
		return null;
	}

}
