/* Copyright (c) 2009, Delft University of Technology

    This file is part of the MEGS package.

    The MEGS pakcage is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    The MEGS pakcage is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the MEGS package.  If not, see
    <http://www.gnu.org/licenses/>.
*/
package mintroute;

import engine.*;

import java.text.DecimalFormat;
import java.util.*;
import java.io.*;

public class MintRouteSystem extends SystemStatus {

	public static final int BASE_STATION_ADDRESS = 126;

	private GlobalPredicate routingCycle;

	private Plot plot;

	public MintRouteSystem() {

		routingCycle = new GlobalPredicate("No routing cycle");
		globalPredicates.add(routingCycle);

		// set the asserts
		assertlist.put(1, "Chosen parent is further from sink than exepected");
		assertlist.put(2, "New parent has created a routing cycle");
		assertlist.put(3, "Message dropped while it would have arrived");
		assertlist
				.put(4,
						"Message is routed to a node further from the sink than the current node");
		assertlist
				.put(5,
						"Message is routed based on outdated inaccurate routing information");
		assertlist.put(6,
				"Message is routed based on a hopcount that is outdated");
		assertlist.put(7, "Information on parent still is outdated after 10s");

		// create the plot
		String[] columns = new String[5];
		columns[0] = "Time";
		columns[1] = "Dead nodes";
		columns[2] = "Avg hopcount";
		columns[3] = "Switches";
		columns[4] = "Avg radio strength";
		plot = new Plot(columns, 0.1, 0);
	}

	@Override
	public void end() {
		super.end();
		try {
			plot.writeData(new File(Settings.PLOT_FILE));
		} catch (IOException e) {
			System.out.println("Error writing plot: " + e);
		}
	}

	protected void updateStatus(Node node, StatusEvent e) {
		MintNode n = (MintNode) node;

		if (e.getName().equals("HOP_COUNT")) {
			n.recordActivity(e.getTime());
			if (n.getHopcount() != e.getValue()) {
				n.setHopcount(e.getValue());
				System.out.println(fullReport());
			}
		} else if (e.getName().equals("PARENT")) {
			n.recordActivity(e.getTime());
			if (n.getParent() != e.getValue()) {
				n.setParent(e.getValue());
				System.out.println(fullReport());
			}
		} else if (e.getName().equals("RADIOSTRENGTH")) {
			n.recordActivity(e.getTime());
			n.setPower(e.getValue());
		} else {
			System.out.println("Unknown status update: " + e);
		}
	}

	@Override
	protected void doAssert(Node n, AssertEvent e) {
		int assertnum = e.getAssertNum();
		switch (assertnum) {
		case 1:
			executeAssert(n, 1);
			executeAssert(n, 2);
			break;
		case 3:
			executeAssert(n, 3);
			break;
		case 4:
			executeAssert(n, 4);
			executeAssert(n, 5);
			executeAssert(n, 6);
			break;
		default:
			errors.add("Unknown assert: " + e);
		}
	}

	@Override
	public MintNode getNode(int i) {
		return (MintNode) super.getNode(i);
	}

	@Override
	public MintNode makeNode(int i) {
		return new MintNode(i, this);
	}

	/**
	 * @return nodes vector list with type MintNode
	 */
	public Vector<MintNode> getAllMintNodes() {
		Vector<MintNode> v = new Vector<MintNode>();
		for (Node n : nodes) {
			v.add((MintNode) n);
		}
		return v;
	}

	@Override
	protected void newTime() {
		routingPredicate();

		// update the plot
		double[] dat = new double[5];
		dat[0] = time; // "Time";
		dat[1] = countDeadNodes(); // "Dead nodes";
		dat[2] = avgHopCount(); // "Avg hopcount";
		dat[3] = avgParentSwitches(); // "Switches";
		dat[4] = avgPower() / 10; // "RadioStrength";
		plot.addPoint(dat);
	}

	/**
	 * Should be called at least at every change of parents, this checks for any
	 * routing cycles and saves it as the predicate
	 */
	private void routingPredicate() {
		boolean cycle = false;
		for (MintNode n : getAllMintNodes()) {
			if (routingCycle(n)) {
				cycle = true;
			}
		}
		// When there is a cycle, the predicate is false!
		routingCycle.updatePredicate(time, !cycle);
	}

	/**
	 * Takes node n and follows the trace to the sink, returns true if a routing
	 * cycle is found Note that if a node has a non existing parent, this will
	 * not be a counted as a routing loop!
	 * 
	 * @param n The start node
	 * @return true iff a routing cycle with n exists
	 */
	private boolean routingCycle(MintNode n) {
		Vector<MintNode> stack = new Vector<MintNode>();

		stack.add(n);
		MintNode node = getNode(n.getParent());
		while (node != null && node.getNumber() != BASE_STATION_ADDRESS) {
			if (stack.contains(node)) {
				// address was already visited earlier
				return true;
			}

			stack.add(node);
			node = getNode(node.getParent());
		}
		return false;
	}

	@Override
	public String fullReport() {
		DecimalFormat df = new DecimalFormat("#.###");
		String s = "Time: " + df.format(time) + " Nodes: " + nodes.size()
				+ " ( " + countDeadNodes() + " dead)\n";
		Object[] ns = nodes.toArray();
		Arrays.sort(ns);

		for (Object o : ns) {
			s += o + "\n";
		}
		s += "Average hopcount: " + avgHopCount() + "\n";
		s += "Average parent switches: " + avgParentSwitches() + "\n";
		s += "Average power setting: " + avgPower() + "\n";
		s += "Amount of nodes not responding: " + countDeadNodes() + "\n";

		return s;
	}

	/**
	 * @param n The node
	 * @param num The number of the assert
	 * @return The value of the assert
	 */
	public boolean checkAssert(Node node, int num)
			throws AssertSkippedException {
		MintNode n = (MintNode) node;
		MintNode parent = null;
		if (n != null) {
			parent = getNode(n.getParent());
		}

		switch (num) {
		case 1:
			if (parent != null) {
				boolean res = n.getHopcount() >= parent.getHopcount();
				if (!res) {
					// assert failed add the extra timed assert
					TimedAssert ass = new TimedAssert(time + 10, n.getNumber(),
							asserts.getAssert(7));
					timedAsserts.add(ass);
				}
				return res;
			} else {
				throw new AssertSkippedException();
			}
		case 2:
			return !routingCycle(n);
		case 3:
			return routingCycle(n);
		case 4:
			if (parent != null) {
				return n.getHopcount() > parent.getHopcount();
			} else {
				throw new AssertSkippedException();
			}
		case 5:
			if (parent != null) {
				return Math.abs(n.getHopcount() - parent.getHopcount()) < 2;
			} else {
				throw new AssertSkippedException();
			}
		case 6:
			if (parent != null) {
				return n.getHopcount() - parent.getHopcount() == 1;
			} else {
				throw new AssertSkippedException();
			}
		case 7:
			if (parent != null) {
				return n.getHopcount() >= parent.getHopcount();
			} else {
				// don't count this as skipped because it is called very often
			}
			return false;
		default:
			errors.add(time + ": Error executing assert with number: " + num);
			throw new AssertSkippedException();
		}
	}

	/**
	 * Returns the avg hopcount for all nodes (ignoring nodes without a parent)
	 * This will return 0 if the amount of nodes with a valid parent is 0
	 */
	public double avgHopCount() {
		double hopcount = 0;
		int hopcount_nodes = 0;
		for (MintNode n : getAllMintNodes()) {
			if (n.getHopcount() != 255) // ignore default values
			{
				hopcount += n.getHopcount();
				hopcount_nodes++;
			}
		}
		if (hopcount_nodes == 0) {
			return 0;
		} else {
			return hopcount / hopcount_nodes;
		}
	}

	/**
	 * Returns the avg amount of parent switches for all nodes (ignoring the
	 * first switch) This will return 0 if the amount of nodes with a valid
	 * parent is 0
	 */
	public double avgParentSwitches() {
		double parentSwitches = 0;
		for (MintNode n : getAllMintNodes()) {
			// ignore the first switch to get a better avg value
			if (n.getParentSwitches() > 1) {

				parentSwitches += n.getParentSwitches() - 1;
			}
		}
		if (nodes.size() == 0) {
			return 0;
		} else {
			return parentSwitches / nodes.size();
		}
	}

	/**
	 * @returns the average power of all nodes
	 */
	public double avgPower() {
		double power = 0;
		for (MintNode n : getAllMintNodes()) {
			power += n.getPower();
		}
		if (nodes.size() == 0) {
			return 0;
		} else {
			return power / nodes.size();
		}
	}
}
